; Master Boot Record image which is simply a program to dump the iBFT
; passed on by an iSCSI booter, for diagnostics.
;
; Copyright (c) 2013 by Kevin McKerron.
; Permission is granted free of charge to distribute, use, and modify this
; software where such use is by the developers of free and open source software
; and for purposes of testing or debugging such software, provided this notice
; is preserved.  For any other use contact krm at fantaplay dotcom.
; THIS PROGRAM IS PROVIDED WITHOUT WARRANTY OF ANY KIND.
;
; BUILDING:
;
; This file can be assembled via most modern versions of Microsoft's MASM (or
; 100% compatible) via the following command:
; ml /c /omf DumpIbft.asm
;
; MASM is generally included with MS Visual Studio.  Check microsoft.com for
; free 'Express' versions of this package and/or freely available Driver
; Development Kits (DDKs) which may include MASM.
;
; Unfortunately, linking is harder, as only a 16-bit version of Microsoft's
; linker will work - modern versions will *not* accept 16-bit object files.
; link.exe v5.60.339 and likely earlier ones will work, via the following
; command:
; link /tiny DumpIbft.obj
;
; NOTE: The resulting file is *not* a normal MS-DOS ".com" file and will crash
; if run from DOS.
;
; A web search should turn up the old linker, but note also that it cannot
; be used directly in a 64-bit OS.
;
; USING:
;
; Write the resulting binary executable file directly into the sector at LBA 0
; of the disk to be booted (this can be done via Linux dd or any number of
; other tools.)  Boot this using the pre-boot solution under test.
;
; A normal BIOS-based PC will respond by displaying iBFT @ <linear-address>
; followed by a hex and ASCII dump of the iBFT, if one is present.  The system
; is then halted so the dump can be examined.  Reboot to a normal OS disk to
; regain use of the system.
;
; Note that the dump intentionally omits lines where the whole paragraph is
; zeros, in order to conserve screen space.  Observe the line addresses in the
; leftmost column.
;
; The display will be set to 50 line mode in order to fit the dump on the
; screen.  If this causes problems with your display you may need to comment out
; that part and reassemble.  If the iBFT is too big to fit on the screen, this
; program won't be of much help without significant changes which might overrun
; the 512 bytes it has to fit within.
;
; NOTE: This program only locates the iBFT via the memory scan method.  It may
; fail to find the iBFT if it was written by the ACPI method, but in practice
; this will usually not be the case anyway.
;
; NOTE: This MBR image does not contain a valid partition table.  Please do not
; use it on a disk with information you wish to keep, at least unless you back
; up the original sector 0 first and replace it prior to invoking any disk
; management tools.


; All subroutines can destroy EAX, EBX, and EDX, even if they are inputs

.model tiny
.code
.386

init_ss			equ		7000h	; we must not write to 80000h or higher

				org		0

				; set up stack (just in case)
_start:			cli
				mov		ax, init_ss
				mov		ss, ax
				mov		sp, 0FFFBh
				sti
				
				; normalize DS to our sector start (just in case)
				call	refpoint
refpoint:		pop		ax
				sub		ax, refpoint
				shr		ax, 4
				mov		dx, cs
				add		ax, dx
				mov		ds, ax
				
				; Find "iBFT" aligned on a para
				mov		di, 8000h
check_seg:		mov		es, di
				cmp		DWord Ptr es:[0], 54464269h
				je		found_sig
next_seg:		inc		di
				jnz		check_seg
			
no_ibft:		mov		si, Offset msg_no_ibft
				call	print_str

halt:			hlt
				jmp		halt
			
found_sig:		; now make sure the checksum passes
				mov		ecx, DWord Ptr es:[4]
				xor		si, si
				xor		dl, dl
@@:				lodsb	es:[si]
				add		dl, al
				dec		ecx
				jnz		@B
				or		dl, dl
				jnz		next_seg
				
				; we have an iBFT
				
				; Say we have it and where
				mov		si, Offset msg_ibft_found
				call	print_str
				mov		ax, es
				mov		dl, 4
				call	print_hex
				mov		si, Offset msg_ibft_suffix
				call	print_str
				
				; Delay so any info from preboot can be seen
				; Don't need to do this now thanks to the video change
				; mov		ah, 86h
				; mov		cx, 396h	; high word - had to make a lot higher on VMware for some reason
				; xor		dx, dx		; low word
				; int		15h
				
				; change video for more lines
				mov		ax, 1112h
				xor		bl, bl
				int		10h

				; Print the iBFT
				
				mov		ax, Word Ptr es:[4]
				mov		remaining_len, ax

dump_para:		; don't dump it if it's all zeros
				xor		ax, ax
				mov		di, line_offset
				push	di
				mov		cx, 8
				repe	scasw
				pop		di
				je		next_para

				mov		ax, di
				mov		dl, 4
				call	print_hex				; offset
				mov		si, Offset dump_offst_sfx
				call	print_str
				
				; dump the paragraph in hex
				call	print_8_hex
				mov		si, Offset dash
				call	print_str
				call	print_8_hex
				
				mov		si, Offset double_space
				call	print_str
				
				; dump ASCII characters
				mov		di, line_offset
				mov		cx, 16
dump_char:		mov		al, es:[di]
				inc		di
				cmp		al, 20h
				jae		@F
				mov		al, '.'
@@:				mov		ah, 0Eh
				xor		bh, bh
				int		10h
				dec		cx
				jnz		dump_char
				
				mov		si, Offset msg_crlf
				call	print_str
				
next_para:		add		line_offset, 10h
				sub		remaining_len, 10h
				ja		dump_para

				jmp		halt

; Variables
				
line_offset		dw		0
remaining_len	dw		?

; Messages

msg_no_ibft		db		"No iBFT!"
msg_crlf		db		0Dh,0Ah,0	; must follow the prior msg
				
msg_ibft_found	db		"iBFT @ ",0
msg_ibft_suffix	db		"0 ",0Dh,0Ah,0

dump_offst_sfx	db		":  ", 0
space			db		" ", 0
dash			db		"-", 0
double_space	db		"  ", 0

hex_digits		db		"0123456789ABCDEF"


; Subroutine to print an ASCIIZ string
;
; Address should be in DS:SI, but SI will be destroyed on return
;
print_str:		mov		al, ds:[si]
				or		al, al
				jnz		@F
				retn
@@:				mov		ah, 0Eh
				xor		bh, bh
				int		10h
				inc		si
				jmp		print_str
				

; Subroutine to print a hex number
;
; Number should be in AX
; Count of digits should be in DL; additional digits are truncated
;
print_hex:		push	si
				
				; alloc buffer to reverse digits
				mov		si, sp
				sub		sp, 6
				
				; null terminator
				dec		si
				mov		Byte Ptr ss:[si], 0
				
				; compute the digits
@@:				mov		bx, ax
				and		bx, 0Fh
				mov		dh, hex_digits[bx]
				dec		si
				mov		ss:[si], dh
				shr		ax, 4
				dec		dl
				jnz		@B
				
				; print the string
				push	ds
				push	ss
				pop		ds
				call	print_str
				pop		ds
				
				add		sp, 6
				pop		si
				retn

; Subroutine to print hex for 8 bytes
;
; ES:DI points to bytes to print, will point past them on return
;
print_8_hex:	push	cx

				mov		cl, 8

@@:				mov		al, es:[di]
				inc		di
				mov		dl, 2
				call	print_hex
				dec		cl
				jz		exit_8_hex
				
				push	si
				mov		si, Offset space
				call	print_str
				pop		si
				
				jmp		@B

exit_8_hex:		pop		cx
				retn

				; MBR bootable signature
				org		1FEh
				db		55h, 0AAh
end _start
; TgrfBp.
